/***************************************************************************************************

  TDispLog  - a C++ Builder 3 class to display and/or log messages to disk.

  Author(s) - Leroy Casterline - casterle@ccltd.com - initial coding.

  Copyright - 1998, Cahill Casterline Limited.
              Permission is granted to use this class in any way you see fit. All I ask is that this
              header not be deleted, that any bug fixes or improvements you may make are sent back
              to me at casterle@ccltd.com, and that you add your name and email address to the
              'authors' list above, along with a brief description of the changes made. Also, if you
              make changes to this class, please update the documentation below.

              I do not claim that this class is designed and/or written as well as it could be, nor
              do I claim that it is bug free. If you use this code, you do so at your own risk! That
              said, I have used this class, and it seems to work.


  Overview
  --------------------------------------------------------------------------------------------------
  I use this class to display and/or log to disk error/debug/other messages. I wrote if for my own
  use, but it seemed useful to me, so I am making it available to anyone who would like to use it. I
  wrote this class using C++ Builder 3.0. As far as I know, it will work under Builder 1.0, but I
  have not tried this.

  I should mention that while I have programmed in C for many years, and Delphi for many fewer, I am
  fairly new to C++. As you look through the code below, you will find places where I've used either
  C standard library calls or the Delphi VCL (such as for file I/O) instead of the C++ methods. I
  don't know if this is good or bad, but it's not C++! I also use AnsiStrings extensively. As this
  class relies on the VCL, I don't think that portability is an issue, but you have been warned.

  The descriptions under Public Functions and Public Properties below are, I hope, sufficient for
  you to make use of the class. I have also included a 'test harness' comprised of the "test.*" and
  "main.*" files included in this distribution. Building and running the 'test' project will allow
  you to play with the various features of this class interactively.

  This class currently uses a TMemo, which it fully manages, to display to the screen. TMemo has
  limitations on the number of strings which it can display, but it is not apparent to me exactly
  what they are. You may set the number of lines the memo will hold using the MaxDispLines property.
  Once the specified number of lines is reached, the oldest line is removed from the memo for each
  new line added. TMemo is not very efficient in doing this, however, and the larger the number of
  lines in the display, the slower the deletions become. You may have to experiment with the number
  of lines to find a value that works for you. The default is 200, which gives marginly acceptable
  performance (your opinion may vary). I have used up to 2000 lines, but performance suffered.

  Logging to disk is much faster than displaying the lines. On my system (P166 w/IDE drives), with
  the property KeepLogOpen set to true, I get about 400 lines/second with full date/time stamping.
  With this property set false, I get about 250 lines/second. This property is used to determine if
  the log file is kept open until the TDispLog object is destroyed, or is opened and closed on every
  call to AddLine(). This is useful in case your program crashes hard, and the log file is empty or
  incomplete because the log was left open and the buffers not completely written to disk. Setting
  KeepLogOpen to false (the default) will prevent this loss of data.

  To use this class, you need to include 'DispLog.h', and add DispLog.h and DispLog.cpp to your
  project.


  Public Functions:
  --------------------------------------------------------------------------------------------------
    void AddLine(AnsiString);                      - write specified line to screen and/or log
    void RemoveDisplay(void);                      - remove the display from screen
    void ClearDisplay(void);                       - clear all lines from memo
    void InitDisplay(TForm *, int Top, int Left, int Height, int Width); - (see notes below)

    Notes: InitDisplay must be called before setting DisplayLine (below) to true. The parameters
           that the InitDisplay call takes are:
             TForm *    - pointer to the form on which the display is to appear
             int Top    - coordinate within form for top of display
             int Left   - coordinate within form for left side of display
             int Height - display height
             int Width  - display width

           If you call RemoveDisplay(), you must call InitDisplay() again before using the display.


  Public Properties:                                                                        defaults
  --------------------------------------------------------------------------------------------------
    unsigned int MaxDispLines - set max number of lines to keep in display                       200
    AnsiString   LogFilename  - name of log file                                         "Debug.Log"
    bool         LogLine      - true to log following lines to disk                            false
    bool         DisplayLine  - true to display following lines to screen                      false
    bool         AppendToLog  - true to append lines to existing log file                       true
    bool         DisplayTime  - true to put time on each line displayed to screen              false
    bool         DisplayMSecs - true to add milliseconds to time displayed                     false
    bool         DisplayDate  - true to put date on each line displayed to screen              false
    bool         LogTime      - true to put time on each line written to disk                  false
    bool         LogMSecs     - true to add milliseconds to time logged                        false
    bool         LogDate      - true to put date on each line written to disk                  false
    unsigned int DisplayDTPos - column position at which date/time is displayed on screen          0
    unsigned int LogDTPos     - column position at which date/time is written to disk              0
    bool         KeepLogOpen  - false to open/close log file for each line logged              false
    bool         DispHorSBar  - true to display a horizontal scroll bar on screen              false
    bool         DispVerSBar  - true to display a vertical scroll bar on screen                false
    AnsiString   SepString    - string used to seperate date/time from logged line             " ~ "

    Notes: If DisplayDTPos or LogDTPos is set to 0, the date and/or time stamp will appear at the
           beginning of the line. A non-zero value will attempt to put the stamp at the specified
           column, at the end of the line. If the line is longer than the specified column number,
           the stamp will be placed at the end of the line, seperated from the line by SepString.

           SepString is a string placed between the line and the date and/or time stamp. It is also
           placed between the date and time if both are selected.

  Exceptions
  --------------------------------------------------------------------------------------------------
  There are six exceptions which may be thrown by this class, all of class Exception.

  If you specify an invalid file name, or for some other reason the log file can't be created or
  opened, an exception is thrown. The Message property will contain the string "File I/O error: "
  followed by the system error string which will explain the error.

  If you set the property DisplayLine to true before you call InitDisplay(), an exception will be
  thrown with the message "Must call InitDisplay() before setting DisplayLine = true!".

  If you set the property KeepLogOpen to either true or false, and the property LogLine is true,
  an exception will be thrown with the message "LogLine must be false before setting KeepLogOpen!".

  If you set the property DisplayDTPos > MAX_POS (currently 200), an exception will be thrown with
  the message "Maximum DisplayDTPos is 200!".

  If you set the property LogDTPos > MAX_POS (currently 200), an exception will be thrown with the
  message "Maximum LogDTPos is 200!".

  If you set the property AppendToLog after writing to the log file or while the property LogToDisk
  is true, an exception will be thrown with the message "AppendToLog must be set before writing to
  log file!"

***************************************************************************************************/
#include <vcl.h>
#include <classes.hpp>
#include <stdio.h>
#pragma hdrstop

#include "DispLog.h"

#pragma package(smart_init)


//==================================================================================================
// TDispLog Constructor - initialize variables
//==================================================================================================
__fastcall TDispLog::TDispLog(void) : TObject()
{
  // be careful when changing these default values. the integrity of the class depends on some
  // of these being set as they are!

  Memo             = NULL;                            // initialize
  LogFileHandle    = -1;                              // "
  LogFileLines     = 0;                               // "
  fMaxDispLines    = 200;                             // "
  fLogFilename     = "Debug.Log";                     // "
  fLogLine         = false;                           // "
  fDisplayLine     = false;                           // "
  fAppendToLog     = true;                            // "
  fDisplayTime     = false;                           // "
  fDisplayMSecs    = false;                           // "
  fDisplayDate     = false;                           // "
  fLogTime         = false;                           // "
  fLogMSecs        = false;                           // "
  fLogDate         = false;                           // "
  fDisplayDTPos    = 0;                               // "
  fLogDTPos        = 0;                               // "
  fKeepLogOpen     = false;                           // "
  fDispHorSBar     = false;                           // "
  fDispVerSBar     = false;                           // "
  fSepString       = " ~ ";                           // "
}



//==================================================================================================
// TDispLog Destructor
//==================================================================================================
__fastcall TDispLog::~TDispLog(void)
{
  if(fKeepLogOpen == true) {                     // log file was kept open
    CloseLogFile();                              // so close it
  }

  if(Memo != NULL) {                             // memo was allocated
    delete Memo;                                 // so free it
  }
}



//==================================================================================================
// InitDisplay - initialize the display memo, set parent, size and max messages
//==================================================================================================
void __fastcall TDispLog::InitDisplay(TForm *ParentForm,   // form on which to display memo
                                      int MemoTop,         // top of memo relative to parent
                                      int MemoLeft,        // left side of memo relative to parent
                                      int MemoHeight,      // height of memo
                                      int MemoWidth)       // width of memo
{
  Memo = new TMemo((TComponent *)NULL);                    // create memo

  Memo->Parent     = ParentForm;                           // show it on the specified form
  Memo->Top        = MemoTop;                              // set location of top of form...
  Memo->Left       = MemoLeft;                             // ...and left of form
  Memo->Width      = MemoWidth;                            // set the form width...
  Memo->Height     = MemoHeight;                           // ...and form height
  Memo->Font->Name = "Courier New";                        // set non-proportional font

  SetScrollBars();                                         // set memo scroll bars
}



//==================================================================================================
// RemoveDisplay - remove the display memo from the parent form
//==================================================================================================
void __fastcall TDispLog::RemoveDisplay(void)
{
  if(Memo != NULL) {                             // memo has been created
    delete Memo;                                 // so remove it
    Memo = NULL;                                 // and null the pointer
  }
}



//==================================================================================================
// ClearDisplay - remove all lines from the parent form
//==================================================================================================
void __fastcall TDispLog::ClearDisplay(void)
{
  if(Memo != NULL) {                             // memo has been created
    Memo->Clear();                               // clear the memo lines
  }
}



//==================================================================================================
// AddLine - add line to log and/or memo
//==================================================================================================
void __fastcall TDispLog::AddLine(AnsiString Msg)
{
  AnsiString MemoLine;                                                    // Msg storage

  if(fDisplayLine == true) {                                              // displaying line
    MemoLine = Msg;                                                       // store message

    if(fDisplayTime || fDisplayDate) {                                    // displaying date or time
      AddDateTime(&MemoLine, fDisplayDate, fDisplayTime, fDisplayMSecs, fDisplayDTPos);
    }

    // although the following BeginUpdate()/EndUpdate() calls cause the memo to flash, they also
    // speed up the deletion process because the screen doesn't have to be updated
    Memo->Lines->BeginUpdate();                                           // stop updating memo

    if(Memo->Lines->Count > (int)fMaxDispLines) {                         // memo full
      Memo->Lines->Delete(0);                                             // delete oldest line
    }

    Memo->Lines->EndUpdate();                                             // resume updating memo

    Memo->Lines->Add(MemoLine);                                           // add line to memo
  }

  if(fLogLine == true) {                                                  // logging line
    MemoLine = Msg;                                                       // store message

    if(fLogTime || fLogDate) {                                            // logging date or time
      AddDateTime(&MemoLine, fLogDate, fLogTime, fLogMSecs, fLogDTPos);   // add date and/or time
    }

    WriteToLog(MemoLine + "\n");                                          // call log function
  }
}



//==================================================================================================
// WriteToLog - write specified string to log file
//==================================================================================================
void __fastcall TDispLog::WriteToLog(AnsiString Msg)
{
  if(fKeepLogOpen == false) {                              // open/close log as needed
    OpenLogFile();                                         // so open it
  }

  try {
    LogFileLines++;                                        // count log file line

    FileWrite(LogFileHandle, Msg.c_str(), Msg.Length());   // write message to file
  } __finally {
    if(fKeepLogOpen == false) {                            // open/close log as needed
      CloseLogFile();                                      // so close it
    }
  }
}



//==================================================================================================
// OpenLogFile - open for append, or create new log file                                   EXCEPTION
//==================================================================================================
void __fastcall TDispLog::OpenLogFile(void)
{
  if(LogFileLines == 0) {                                    // writing 1st line of this program run
    if(fAppendToLog) {                                       // appending to existing log
      if(FileExists(fLogFilename)) {                         // file exists...
        LogFileHandle = FileOpen(fLogFilename, fmOpenWrite); // ...append to existing
      } else {                                               // file doesn't exist...
        LogFileHandle = FileCreate(fLogFilename);            // ...so create it
      }

      if(LogFileHandle >= 0) {                               // success
        FileSeek(LogFileHandle, 0, 2);                       // move to the end of the file
      }
    } else {                                                 // overwrite existing log
      LogFileHandle = FileCreate(fLogFilename);              // create new file, overwrite existing
    }
  } else {                                                   // not writing 1st line, always append
    LogFileHandle = FileOpen(fLogFilename, fmOpenWrite);     // append to existing, create if none

    if(LogFileHandle >= 0) {                                 // success
      FileSeek(LogFileHandle, 0, 2);                         // move to the end of the file
    }
  }

  if(LogFileHandle < 0) {                                    // file open problem
    throw Exception("File I/O error:" + GetSystemErrorMessage());
  }
}



//==================================================================================================
// CloseLogFile - close the log file if it is open, reset file handle
//==================================================================================================
void __fastcall TDispLog::CloseLogFile(void)
{
  if(LogFileHandle >= 0) {                       // log file open
    FileClose(LogFileHandle);                    // so close it

    LogFileHandle = -1;                          // set file handle to 'closed'
  }
}



//==================================================================================================
// GetSystemErrorMessage - return string containing Win32 'GetLastError()' message
//==================================================================================================
AnsiString __fastcall TDispLog::GetSystemErrorMessage(void)
{
  LPVOID lpMsgBuf;                                                // pointer to allocated message

  FormatMessage(                                                  // get 'GetLastError' message
    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),                    // default language
    (LPTSTR) &lpMsgBuf,
    0,
    NULL
  );

  AnsiString Buffer = (char *)lpMsgBuf;                           // store in AnsiString

  LocalFree(lpMsgBuf);                                            // free auto-allocated buffer

  return Buffer;                                                  // return our error message
}



//==================================================================================================
// AddDateTime - pad line, add date and time stamps, as specified
//==================================================================================================
void __fastcall TDispLog::AddDateTime(AnsiString *Msg,bool AddDate,bool AddTime,bool AddMS, int Pos)
{
  unsigned short Hour;                                     // decoded hour value
  unsigned short Minute;                                   // decoded minute value
  unsigned short Second;                                   // decoded second value
  unsigned short mSecond;                                  // decoded millisecond value
  char           TimeBuf[30];                              // formatted time buffer
  TDateTime      DT = Now();                               // get, store date and time
  AnsiString     DateFmt = "mm/dd/yyyy";                   // date format string
  AnsiString     TimeFmt = "hh:nn:ss";                     // time format string


  if(AddDate == true || AddTime == true) {                 // adding date or time
    if(Pos == 0) {                                         // date/time at start of line
      AnsiString Buffer = "";                              // line buffer

      if(AddDate == true) {                                // date wanted
        Buffer += DT.FormatString(DateFmt) + fSepString;   // put date in buffer
      }

      if(AddTime == true) {                                // time wanted
        if(AddMS == true) {                                // add milliseconds to time
          DecodeTime(DT, Hour, Minute, Second, mSecond);   // decode current time

          sprintf(TimeBuf, "%2.2d:%2.2d:%2.2d.%3.3d", Hour, Minute, Second, mSecond);

          Buffer += TimeBuf;                               // add formatted time to line
        } else {
          Buffer += DT.FormatString(TimeFmt);              // format time w/o milliseconds
        }

        Buffer = Buffer + fSepString;                      // add seperating string
      }

      Buffer += *Msg;                                      // add rest of line to buffer
      *Msg    = Buffer;                                    // then store whole thing back in msg
    } else {                                               // date/time at end of line
      PadString(Msg, Pos);                                 // pad string

      *Msg += fSepString;                                  // add string seperating date/time

      if(AddDate == true) {                                // if date is to be added
        *Msg += Now().FormatString(DateFmt);               // add formatted date
      }

      if(AddTime == true) {                                // if time to be added
        if(AddDate == true) {                              // and date was added
          *Msg += fSepString;                              // add seperating space
        }

        if(AddMS == true) {                                // add milliseconds to time
          DecodeTime(DT, Hour, Minute, Second, mSecond);   // get current time

          sprintf(TimeBuf, "%2.2d:%2.2d:%2.2d.%3.3d", Hour, Minute, Second, mSecond);

          *Msg += TimeBuf;                                 // add time w/msecs to line
        } else {
          *Msg += DT.FormatString(TimeFmt);                // format time w/o milliseconds
        }
      }
    }
  }
}



//==================================================================================================
// PadString - pad the specified AnsiString w/spaces to the specified length
//==================================================================================================
void __fastcall TDispLog::PadString(AnsiString *Msg, int Len)
{
  char PadBuf[MAX_POS + 1];                            // buffer for padding
  int Pad = Len - Msg->Length() - 1;                   // get padding length

  if(Pad > 0) {                                        // padding needed
    memset((void *)PadBuf, ' ', Pad);                  // fill w/spaces

    PadBuf[Pad] = 0;                                   // null terminate

    *Msg += PadBuf;                                    // add padding
  }
}



//==================================================================================================
// SetScrollBars - set the memo scroll bars
//==================================================================================================
void __fastcall TDispLog::SetScrollBars(void)
{
  if(Memo != NULL) {                                       // memo has been created
    if(fDispHorSBar == true && fDispVerSBar == true) {     // display both scroll bars
      Memo->ScrollBars = ssBoth;
    } else if(fDispHorSBar == true) {                      // display horizontal scroll bar
      Memo->ScrollBars =  ssHorizontal;
    } else if(fDispVerSBar == true) {                      // display vertical scroll bar
      Memo->ScrollBars = ssVertical;
    } else {                                               // display no scroll bars
      Memo->ScrollBars = ssNone;
    }
  }
}



//==================================================================================================
//==================================================================================================
//==================================================================================================
// property get/set functions
//==================================================================================================
//==================================================================================================
//==================================================================================================



//==================================================================================================
// SetLogFilename - set the name of the log file
//==================================================================================================
void __fastcall TDispLog::SetLogFilename(AnsiString Name)
{
  fLogFilename = Name;                           // store file name

  if(fLogLine == true && fKeepLogOpen == true) { // already logging lines w/file kept open
    CloseLogFile();                              // so close the file
    OpenLogFile();                               // and reopen it with the new name
  }
}



//==================================================================================================
// SetLogLine - set the log-line-to-disk flag to the specified value
//==================================================================================================
void __fastcall TDispLog::SetLogLine(bool Flag)
{
  if(fLogLine != Flag) {                         // ignore subsequent same calls
    fLogLine = Flag;                             // store flag value

    if(fKeepLogOpen == true) {                   // keeping log open
      if(fLogLine == true) {                     // starting log
        OpenLogFile();                           // so open log file
      } else {                                   // stopping log
        CloseLogFile();                          // so close log file
      }
    }
  }
}



//==================================================================================================
// SetDisplayLine - set the display-line-to-memo flag to the specified value               EXCEPTION
//==================================================================================================
void __fastcall TDispLog::SetDisplayLine(bool Flag)
{
  if(Flag == true) {                             // starting display
    if(Memo != NULL) {                           // memo has been created
      fDisplayLine = true;                       // store flag value
    } else {                                     // memo has NOT been created
      throw Exception("Must call InitDisplay() before setting DisplayLine = true!");
    }
  } else {                                       // stopping display
    fDisplayLine = false;                        // set flag false
  }
}



//==================================================================================================
// SetAppendToLog - set the append-to-log flag to specified value                          EXCEPTION
//==================================================================================================
void __fastcall TDispLog::SetAppendToLog(bool Flag)
{
  if(fLogLine == true || LogFileLines > 0) {     // file open or already written to
    throw Exception("AppendToLog must be set before writing to log file!");
  } else {
    fAppendToLog = Flag;                         // store flag value
  }
}



//==================================================================================================
// SetDisplayDTPos - set the character position of the date and/or time in the memo line   EXCEPTION
//==================================================================================================
void __fastcall TDispLog::SetDisplayDTPos(unsigned int Value)
{
  if(Value > MAX_POS) {                          // too long
    throw Exception("Maximum DisplayDTPos is " + IntToStr(MAX_POS) + "!");
  } else {
    fDisplayDTPos = Value;                       // store value
  }
}



//==================================================================================================
// SetLogDTPos - set the character position of the date and/or time in the log line        EXCEPTION
//==================================================================================================
void __fastcall TDispLog::SetLogDTPos(unsigned int Value)
{
  if(Value > MAX_POS) {                          // too long
    throw Exception("Maximum LogDTPos is "  + IntToStr(MAX_POS) + "!");
  } else {
    fLogDTPos = Value;                           // store value
  }
}



//==================================================================================================
// SetKeepLogOpen - set the keep-log-open flag to the specified value                      EXCEPTION
//==================================================================================================
void __fastcall TDispLog::SetKeepLogOpen(bool Flag)
{
  if(fLogLine == true) {                        // already logging to disk
    throw Exception("LogLine must be false before setting KeepLogOpen!");
  }

  fKeepLogOpen = Flag;                           // store flag value
}



//==================================================================================================
// SetDispHorSBar - set the display-horizontal-scroll-bar flag to the specified value
//==================================================================================================
void __fastcall TDispLog::SetDispHorSBar(bool Flag)
{
  fDispHorSBar = Flag;                           // store flag value

  SetScrollBars();                               // update memo to reflect new flag values
}



//==================================================================================================
// SetDispVerSBar - set the display-vertical-scroll-bar flag to the specified value
//==================================================================================================
void __fastcall TDispLog::SetDispVerSBar(bool Flag)
{
  fDispVerSBar = Flag;                           // store flag value

  SetScrollBars();                               // update memo to reflect new flag value
}




